use prost_reflect::{
    prost_types::{
        field_descriptor_proto::Type, DescriptorProto, FieldDescriptorProto, FileDescriptorProto,
    },
    DescriptorPool, MessageDescriptor,
};
use pyo3::{exceptions::PyRuntimeError, PyAny, PyResult};
use std::sync::{Mutex, OnceLock};

use crate::py_any_extras::PyAnyExtras;

pub fn create_cached_descriptor(obj: &PyAny) -> PyResult<MessageDescriptor> {
    static DESCRIPTOR_POOL: OnceLock<Mutex<DescriptorPool>> = OnceLock::new();
    let mut pool = DESCRIPTOR_POOL
        .get_or_init(|| Mutex::new(DescriptorPool::new()))
        .lock()
        .unwrap();

    create_cached_descriptor_in_pool(obj, &mut pool)
}

fn create_cached_descriptor_in_pool<'py>(
    obj: &'py PyAny,
    pool: &mut DescriptorPool,
) -> PyResult<MessageDescriptor> {
    let name = obj.qualified_class_name()?;
    if let Some(desc) = pool.get_message_by_name(&name) {
        return Ok(desc);
    }

    let meta = obj.get_proto_meta()?;

    let mut message = DescriptorProto {
        name: Some(name.clone()),
        ..Default::default()
    };

    let mut file = FileDescriptorProto {
        name: Some(name.clone()),
        ..Default::default()
    };

    for item in meta
        .getattr("meta_by_field_name")?
        .call_method0("items")?
        .iter()?
    {
        let (field_name, field_meta) = item?.extract::<(&str, &'py PyAny)>()?;
        message.field.push({
            let mut field = FieldDescriptorProto {
                name: Some(field_name.to_string()),
                number: Some(field_meta.getattr("number")?.extract::<i32>()?),
                ..Default::default()
            };
            field.set_type(map_type(
                field_meta.getattr("proto_type")?.extract::<&str>()?,
            )?);

            if field.r#type() == Type::Message {
                let instance = meta.create_instance(field_name)?;
                let cls_name = instance.qualified_class_name()?;
                field.type_name = Some(cls_name.to_string());
                create_cached_descriptor_in_pool(instance, pool)?;
            }

            field
        });
    }

    file.message_type.push(message);

    pool.add_file_descriptor_proto(file)
        .map_err(|_| PyRuntimeError::new_err("Error on proto registration"))?;

    Ok(pool.get_message_by_name(&name).expect("Just registered..."))
}

fn map_type(str: &str) -> PyResult<Type> {
    match str {
        "enum" => Ok(Type::Enum),
        "bool" => Ok(Type::Bool),
        "int32" => Ok(Type::Int32),
        "int64" => Ok(Type::Int64),
        "uint32" => Ok(Type::Uint32),
        "uint64" => Ok(Type::Uint64),
        "sint32" => Ok(Type::Sint32),
        "sint64" => Ok(Type::Sint64),
        "float" => Ok(Type::Float),
        "double" => Ok(Type::Double),
        "fixed32" => Ok(Type::Fixed32),
        "sfixed32" => Ok(Type::Sfixed32),
        "fixed64" => Ok(Type::Fixed64),
        "sfixed64" => Ok(Type::Sfixed64),
        "string" => Ok(Type::String),
        "bytes" => Ok(Type::Bytes),
        "message" => Ok(Type::Message),
        _ => Err(PyRuntimeError::new_err("Unsupported type")),
    }
}
